from cosmic_toolkit.message_bus import MessageBus
from cosmic_toolkit.models import AggregateRoot, Entity, Event
from cosmic_toolkit.repository import AbstractRepository
from cosmic_toolkit.unit_of_work import BaseUnitOfWork

__all__ = [
    "AggregateRoot",
    "AbstractRepository",
    "BaseUnitOfWork",
    "Entity",
    "Event",
    "MessageBus",
]



import logging
from inspect import Parameter, signature
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from cosmic_toolkit.models import Event

logger = logging.getLogger(__name__)


class MessageBus:
    def __init__(
        self,
        handlers: Dict[Type[Event], List[Callable]],
        ignore_missing_handlers: Optional[bool] = False,
        unit_of_work_kwarg_name: Optional[str] = "uow",
        **dependencies,
    ):
        self._dependencies = dependencies
        self._handlers = handlers
        # If True, RuntimeError will not be raised if there are no handlers for an
        # event
        self._ignore_missing_handlers = ignore_missing_handlers
        self._unit_of_work_kwarg_name = unit_of_work_kwarg_name

    @property
    def dependencies(self) -> Dict[str, Any]:
        return self._dependencies

    @staticmethod
    def _cull_dependencies(
        parameters: Dict[str, Parameter], dependencies: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Find arguments for a handler using signature parameters. Skips the
        event argument since that's expected to be the first argument and is explicitly
        passed by _handle_event()"""
        return {
            name: dependencies.get(name)
            for name, param in parameters.items()
            if name != "event"
        }

    def _get_handlers_for_event(self, event: Event) -> List[Callable]:
        """Return the first list of handlers found for event.
        Searches for handlers using event type and parents' types."""

        for klass in [event.__class__] + [t for t in event.__class__.__bases__]:
            handlers = self._handlers.get(klass)

            if handlers:
                return handlers

        raise RuntimeError(f"No handlers found for {event.__class__}")

    def _resolve_dependencies(
        self, handler: Callable, **dependencies
    ) -> Tuple[list, dict]:
        # Combine all dependencies
        all_provided_deps = {**self._dependencies, **dependencies}

        # Bind dependencies
        sig = signature(handler)
        culled_deps = self._cull_dependencies(sig.parameters, all_provided_deps)
        bounded = sig.bind_partial(**culled_deps)

        # Return args + kwargs
        return list(bounded.args), bounded.kwargs

    async def _handle_event(self, event: Event, **dependencies) -> List[Event]:
        events = []
        handlers = []

        try:
            # This will raise RuntimeError if there are no handlers for the event
            # _ignore_missing_handlers dictates the functionality when this happens
            # Default is to raise RuntimeError but event can be ignored by instantiating
            # message bus with ignore_missing_handlers=True
            handlers = self._get_handlers_for_event(event)
        except RuntimeError:
            if not self._ignore_missing_handlers:
                raise

        for handler in handlers:
            args, kwargs = self._resolve_dependencies(handler, **dependencies)

            logger.debug("Using %s to handle %s", handler, event)
            await handler(event, *args, **kwargs)

            # Attempt to collect new events published by handler
            # We need a Unit of Work dependency for this
            # Not all use cases require UoW, so if UoW isn't in deps,
            # it's not an issue
            uow = {**self._dependencies, **dependencies}.get(
                self._unit_of_work_kwarg_name
            )

            if uow:
                events.extend(uow.collect_new_events())

        return events

    def add_dependencies(self, **dependencies):
        self._dependencies.update(dependencies)

    async def handle(self, event: Event, **dependencies):
        queue = [event]

        # Domain models can publish new events which is why we use a queue here
        while queue:
            new_events = await self._handle_event(queue.pop(0), **dependencies)
            queue.extend(new_events)



from abc import ABCMeta, abstractmethod
from typing import Dict, Generator, List

from cosmic_toolkit.models import Event
from cosmic_toolkit.repository import AbstractRepository


class BaseUnitOfWork(metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        """Instantiate Unit of Work - arguments are passed into constructors of
        repositories"""
        self._args = args
        self._kwargs = kwargs
        self._repositories: Dict[str, AbstractRepository] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        cls._repository_classes = {
            k: v for k, v in kwargs.items() if issubclass(v, AbstractRepository)
        }

    async def __aenter__(self) -> "BaseUnitOfWork":
        # Instantiate repositories if they haven't been instantiated
        if not self._repositories:
            self._repositories = {
                k: v(*self._args, **self._kwargs)
                for k, v in self._repository_classes.items()
            }

        return self

    async def __aexit__(self, exc_type, exc, tb):
        # If transaction is committed, rollback shouldn't error
        # This is here as a fallback
        await self.rollback()

    def __eq__(self, other: "BaseUnitOfWork") -> bool:
        return self.__repr__() == other.__repr__()

    def __getattr__(self, item: str) -> AbstractRepository:
        # Enable accessing repositories as attributes
        # E.g. uow.customers.add()
        try:
            return self._repositories[item]
        except KeyError:
            raise AttributeError(
                f"{self.__class__.__name__} does not have {item!r} repository"
            )

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}, "
            f"repositories={[r.__repr__() for r in self._repositories.values()]}>"
        )

    def collect_new_events(self) -> Generator[List[Event], None, None]:
        for repository in self._repositories.values():
            for entity in repository.seen:
                yield from entity.events

    @abstractmethod
    async def commit(self):
        ...

    @abstractmethod
    async def rollback(self):
        ...



import pytest

from cosmic_toolkit import AbstractRepository, AggregateRoot

pytestmark = pytest.mark.asyncio


def test_abstract_repository_incorrect_entity_type():
    class Item:
        ...

    with pytest.raises(TypeError) as e:

        class ItemRepository(AbstractRepository, entity_type=Item):
            async def _add(self, entity: AggregateRoot):
                ...

            async def _get(self, id: str) -> AggregateRoot:
                ...

            async def _update(self, entity: AggregateRoot):
                ...

    assert str(e.value) == "Entity must inherit from AggregateRoot"


async def test_abstract_repository_add_update_enforces_entity(
    test_entities, test_repositories
):
    entity_a = test_entities["EntityA"].init("hello")
    entity_b = test_entities["EntityB"].init("world")
    repository_a = test_repositories["TestRepositoryA"]()

    await repository_a.add(entity_a)

    with pytest.raises(TypeError) as e:
        await repository_a.add(entity_b)

    assert str(e.value) == "Expecting entity of type EntityA"
    assert list(repository_a.seen)[0] == entity_a


def test_abstract_repository_subclass_kwargs(test_entities):
    """Ensure that class-level keyword arguments are captured and stored so they can
    be used by subclasses"""

    class ARepository(
        AbstractRepository,
        entity_type=test_entities["EntityA"],
        collection_name="test",
    ):
        async def _add(self, entity: AggregateRoot):
            ...

        async def _get(self, id: str) -> AggregateRoot:
            ...

        async def _update(self, entity: AggregateRoot):
            ...

    a_repository = ARepository()

    assert a_repository._init_kwargs == {"collection_name": "test"}



import inspect
from abc import ABCMeta, abstractmethod
from typing import Any, Generator, List, Optional, Type

from orjson import dumps
from pydantic import BaseModel

from cosmic_toolkit.types import JSONSerializer, NormalDict


class Event(BaseModel):
    ...


class AggregateRoot:
    _events: List[Event]

    def __init__(self, *args, **kwargs):
        self._events = []

    @property
    def events(self) -> Generator[Event, None, None]:
        while self._events:
            yield self._events.pop(0)

    def _add_event(self, event: Event):
        self._events.append(event)


class DefaultSerializer:
    def __call__(self, obj: Any):
        raise TypeError


class Entity(metaclass=ABCMeta):
    def __eq__(self, other: "Entity") -> bool:
        return self.dict() == other.dict()

    def __hash__(self) -> int:
        # Making entities hashable enables using sets which is necessary for
        # BaseUnitOfWork to collect domain events
        return hash(self.json())

    def __init_subclass__(
        cls, default_json_serializer: Optional[Type[JSONSerializer]] = None, **kwargs
    ):
        cls._default_json_serializer = (
            default_json_serializer if default_json_serializer else DefaultSerializer
        )

    def __repr__(self) -> str:
        """Create entity representation. Searches for entity properties to create
        repr"""
        props = []

        # The next few lines feel hacky but there doesn't seem to be a clear way to
        # find an object's properties.
        # To ensure that properties are found, be sure that @property getters
        # match argument names in __init__().
        #
        # E.g.
        #
        # class User:
        #     def __init__(self, id: int, name: str, is_active: bool):
        #         ...
        #
        #     @property
        #     def id(self) -> int:
        #         ...
        #
        #     @property
        #     def name(self) -> str:
        #         ...
        #
        # >>> user.__repr__()
        # User(id=987, name='Random User')
        constructor_param_names = inspect.signature(self.__init__).parameters.keys()

        for p in dir(self):
            if not callable(p) and p in constructor_param_names:
                props.append(f"{p}={getattr(self, p)!r}")

        return f"{self.__class__.__name__}({', '.join(props)})"

    @classmethod
    @abstractmethod
    def init(cls, *args, **kwargs) -> "Entity":
        ...

    @abstractmethod
    def dict(self) -> NormalDict:
        ...

    def json(self) -> bytes:
        return dumps(self.dict(), default=self._default_json_serializer())

    class DoesNotExist(Exception):
        ...



from abc import ABCMeta, abstractmethod
from typing import Any, Type

from cosmic_toolkit.models import AggregateRoot


class AbstractRepository(metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        self.seen = set()

    def __init_subclass__(cls, entity_type: Type[AggregateRoot], **kwargs):
        if not issubclass(entity_type, AggregateRoot):
            raise TypeError(f"Entity must inherit from {AggregateRoot.__name__}")

        cls._entity_type = entity_type
        cls._init_kwargs = kwargs

    def __repr__(self):
        return f"<{self.__class__.__name__}, entity_type={self._entity_type.__name__}>"

    def _check_entity_type(self, entity):
        if not type(entity) == self._entity_type:
            raise TypeError(f"Expecting entity of type {self._entity_type.__name__}")

    async def add(self, entity: AggregateRoot):
        self._check_entity_type(entity)

        await self._add(entity)
        self.seen.add(entity)

    async def get(self, id: Any) -> AggregateRoot:
        entity = await self._get(id)

        if entity:
            self.seen.add(entity)

        return entity

    async def update(self, entity):
        self._check_entity_type(entity)

        await self._update(entity)
        self.seen.add(entity)

    @abstractmethod
    async def _add(self, entity: AggregateRoot):
        ...

    @abstractmethod
    async def _get(self, id: str) -> AggregateRoot:
        ...

    @abstractmethod
    async def _update(self, entity: AggregateRoot):
        ...



from typing import List, Optional, Tuple

import pytest

from cosmic_toolkit import (
    AbstractRepository,
    AggregateRoot,
    BaseUnitOfWork,
    Entity,
    Event,
    MessageBus,
)
from cosmic_toolkit.types import NormalDict

pytestmark = pytest.mark.asyncio


class TelemetryReceived(Event):
    message: str


class Telemetry(AggregateRoot):
    def __init__(self, id: int, message: str):
        super().__init__()

        self._id = id
        self._message = message

    @classmethod
    def init(cls, message: str, id: Optional[int] = None) -> "Telemetry":
        if not id:
            id = hash(message)

        return cls(id, message)

    @property
    def id(self) -> int:
        return self._id

    @property
    def message(self) -> str:
        return self._message

    def dict(self) -> NormalDict:
        return {
            "id": self._id,
            "message": self._message,
        }


class TelemetryRepository(AbstractRepository, entity_type=Telemetry):
    def __init__(self):
        super().__init__()

        self._items = {}

    async def _add(self, entity: Telemetry):
        self._items[hash(entity.message)] = entity

    async def _get(self, id: str) -> Telemetry:
        return self._items[id]

    async def _update(self, entity: Telemetry):
        return self._add(entity)


class UnitOfWork(BaseUnitOfWork, telemetry=TelemetryRepository):
    async def commit(self):
        ...

    async def rollback(self):
        ...


Log = List[Tuple[int, str]]


class TelemetryLog:
    _log: Log

    def __init__(self):
        self._log = []

    @property
    def log(self) -> Log:
        return self._log

    def add(self, id: int, message: str):
        self._log.append((id, message))


async def record_telemetry(event: TelemetryReceived, uow: BaseUnitOfWork):
    async with uow:
        point = Telemetry.init(event.message)

        await uow.telemetry.add(point)
        await uow.commit()


async def update_log(event: TelemetryReceived, uow: BaseUnitOfWork, log: TelemetryLog):
    telemetry_id = hash(event.message)

    async with uow:
        point = await uow.telemetry.get(telemetry_id)

    log.add(telemetry_id, point.message)


async def test_message_bus_handle():
    log = TelemetryLog()
    uow = UnitOfWork()

    message_bus = MessageBus(
        {
            TelemetryReceived: [record_telemetry, update_log],
        },
        log=log,
        uow=uow,
    )
    event_a = TelemetryReceived(message="test123")
    event_a_id = hash(event_a.message)

    # Publish message to bus and expect it to be handled
    await message_bus.handle(event_a)

    # Check that telemetry was saved
    expected_data_point = await uow.telemetry.get(event_a_id)
    data = expected_data_point.dict()

    assert data["id"] == event_a_id
    assert data["message"] == event_a.message

    # Publish another message
    event_b = TelemetryReceived(message="hello_world")
    event_b_id = hash(event_b.message)

    await message_bus.handle(event_b)

    # Check that log was updated
    assert len(log.log) == 2
    assert log.log[0] == (event_a_id, event_a.message)
    assert log.log[1] == (event_b_id, event_b.message)


def test_message_bus_add_dependencies():
    log = TelemetryLog()
    uow = UnitOfWork()

    message_bus = MessageBus({}, uow=uow)

    assert len(message_bus.dependencies.values()) == 1
    assert message_bus.dependencies["uow"] == UnitOfWork()

    # Now add a dependency
    message_bus.add_dependencies(log=log)

    assert len(message_bus.dependencies.values()) == 2
    assert message_bus.dependencies["log"] == log


async def test_message_bus_handle_ignore_missing_handler():
    message_bus = MessageBus({}, ignore_missing_handlers=True)
    event_a = TelemetryReceived(message="test123")

    # Nothing happens
    await message_bus.handle(event_a)


async def test_message_bus_handle_raise_error_on_missing_handler():
    # By default, ignore_missing_handlers=False
    message_bus = MessageBus({})
    event_a = TelemetryReceived(message="test123")

    with pytest.raises(RuntimeError):
        await message_bus.handle(event_a)



from typing import Dict, Type

import pytest

from cosmic_toolkit import (
    AbstractRepository,
    AggregateRoot,
    BaseUnitOfWork,
    Entity,
    Event,
)
from cosmic_toolkit.types import NormalDict


class EntityA(AggregateRoot, Entity):
    def __init__(self, a: str):
        super().__init__()
        self._a = a

    @classmethod
    def init(cls, a: str) -> "AggregateRoot":
        return cls(a)

    @property
    def a(self) -> str:
        return self._a

    def dict(self) -> NormalDict:
        return {"a": self._a}


class EntityB(AggregateRoot, Entity):
    def __init__(self, a: str):
        super().__init__()
        self._a = a

    @classmethod
    def init(cls, a: str) -> "AggregateRoot":
        return cls(a)

    @property
    def a(self) -> str:
        return self._a

    def dict(self) -> NormalDict:
        return {"a": self._a}


class ATriggered(Event):
    ...


class BTriggered(Event):
    ...


class TestRepositoryA(AbstractRepository, entity_type=EntityA):
    async def _add(self, entity: EntityA):
        ...

    async def _get(self, id: str) -> EntityA:
        ...

    async def _update(self, entity: EntityA):
        ...


class TestRepositoryB(AbstractRepository, entity_type=EntityB):
    async def _add(self, entity: EntityA):
        ...

    async def _get(self, id: str) -> EntityA:
        ...

    async def _update(self, entity: EntityA):
        ...


class TestUnitOfWork(
    BaseUnitOfWork,
    a_items=TestRepositoryA,
    b_items=TestRepositoryB,
):
    async def commit(self):
        ...

    async def rollback(self):
        ...


@pytest.fixture
def test_entities() -> Dict[str, Type[Entity]]:
    return {
        "EntityA": EntityA,
        "EntityB": EntityB,
    }


@pytest.fixture
def test_events() -> Dict[str, Type[Event]]:
    return {
        "ATriggered": ATriggered,
        "BTriggered": BTriggered,
    }


@pytest.fixture
def test_repositories() -> Dict[str, Type[TestRepositoryA]]:
    return {
        "TestRepositoryA": TestRepositoryA,
        "TestRepositoryB": TestRepositoryB,
    }


@pytest.fixture
def test_unit_of_work() -> Type[TestUnitOfWork]:
    return TestUnitOfWork



from setuptools import setup, find_packages

VERSION = "0.4.1"


with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="cosmic-toolkit",
    version=VERSION,
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=[
        "orjson==3.4.3",
        "pydantic==1.7.1",
    ],
    packages=find_packages(),
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="cosmic,messagebus,domain driven design,clean architecture"
)



import pytest

from cosmic_toolkit import AbstractRepository

pytestmark = pytest.mark.asyncio


async def test_base_unit_of_work_repository_attributes(test_unit_of_work):
    uow = test_unit_of_work()

    # Ensure repos can be accessed as attributes, but only after the uow is used as a
    # context manager
    assert not hasattr(uow, "a_items")
    assert not hasattr(uow, "b_items")

    with pytest.raises(AttributeError) as e:
        uow.a_items

    assert str(e.value) == "TestUnitOfWork does not have 'a_items' repository"

    # Now we can access repositories
    async with uow:
        assert hasattr(uow, "a_items")
        assert hasattr(uow, "b_items")

        assert isinstance(uow.a_items.seen, set)

    # uow has been used as context manager so now we expect repositories to be
    # instantiated
    assert isinstance(uow.a_items, AbstractRepository)


async def test_base_unit_of_work_collect_new_events(
    test_entities, test_events, test_unit_of_work
):
    uow = test_unit_of_work()

    # Add entities
    async with uow:
        entity_a = test_entities["EntityA"].init("hello")
        entity_a._add_event(test_events["ATriggered"]())
        entity_a._add_event(test_events["BTriggered"]())
        entity_b = test_entities["EntityB"].init("world")
        entity_b._add_event(test_events["BTriggered"]())

        await uow.a_items.add(entity_a)
        await uow.b_items.add(entity_b)

        await uow.commit()

    # Collect events
    events = list(uow.collect_new_events())

    assert len(events) == 3
    assert isinstance(events[0], test_events["ATriggered"])
    assert isinstance(events[1], test_events["BTriggered"])
    assert isinstance(events[2], test_events["BTriggered"])

    # Events should be cleared out
    assert len(list(uow.collect_new_events())) == 0



from datetime import datetime
from random import randint
from typing import Any, List, Optional, Tuple
from uuid import UUID, uuid4

import pytest

from cosmic_toolkit import AggregateRoot, Entity, Event
from cosmic_toolkit.types import NormalDict

# Aggregates consist of one or more child entities
# We need an aggregate root to create an aggregate
# Aggregates capture domain events


class ConstraintViolation(Exception):
    ...


class ValidationError(Exception):
    ...


class Suite(Entity):
    def __init__(self, number: str, name: str, floor: int, sqft: int, leased: bool):
        self._floor = floor
        self._leased = leased
        self._name = name
        self._number = number
        self._sqft = sqft

    @classmethod
    def init(
        cls,
        number: str,  # str because "number" can include/be letters (e.g. 201B)
        name: str,
        floor: int,
        sqft: int,
        leased: bool,
    ) -> "Suite":
        # Ensure floor makes sense
        if not -25 < floor < 125:
            raise ValueError(f"Floor number {floor} is invalid")

        return cls(
            number,
            name,
            floor,
            sqft,
            leased,
        )

    @property
    def leased(self) -> bool:
        return self._leased

    @property
    def number(self) -> str:
        return self._number

    @property
    def sqft(self) -> int:
        return self._sqft

    def dict(self) -> NormalDict:
        return {
            "name": self._name,
            "number": self._number,
            "floor": self._floor,
            "sqft": self._sqft,
            "leased": self._leased,
        }

    def lease(self):
        if self.leased:
            raise ConstraintViolation("Suite is already leased")

        self._leased = True

    def make_available(self):
        if not self.leased:
            raise ConstraintViolation("Suite is already available")

        self._leased = False


class Address:
    """Address value object"""

    def __init__(
        self,
        street_line_1: str,
        locality: str,
        region: str,
        postal_code: str,
        country_code_iso_alpha_3: str,
        street_line_2: Optional[str] = None,
    ):
        if len(country_code_iso_alpha_3) != 3:
            raise ValueError("Country code is expected to be of len")

        country_code_iso_alpha_3 = country_code_iso_alpha_3.upper()

        if country_code_iso_alpha_3 not in ["AUS", "CAN", "CYM", "LCA", "RWA", "USA"]:
            raise ValueError(f"Country {country_code_iso_alpha_3!r} is not supported")

        self._country_code = country_code_iso_alpha_3
        self._locality = locality
        self._postal_code = postal_code
        self._region = region
        self._street_line_1 = street_line_1
        self._street_line_2 = street_line_2

    def dict(self) -> NormalDict:
        return {
            "street_line_1": self._street_line_1,
            "street_line_2": self._street_line_2,
            "locality": self._locality,
            "region": self._region,
            "postal_code": self._postal_code,
            "country_code_iso_alpha_3": self._country_code,
        }


class SuiteAdded(Event):
    building_id: UUID
    number: str


class SuiteLeased(SuiteAdded):
    ...


class SuiteMadeAvailable(SuiteAdded):
    ...


class SuiteRemoved(Event):
    building_id: UUID
    name: str
    number: str


class Building(AggregateRoot, Entity):
    def __init__(self, id: UUID, name: str, address: Address, suites: List[Suite]):
        super().__init__()
        self._id = id
        self._address = address
        self._name = name
        self._suites = {s.number: s for s in suites}

    @classmethod
    def init(
        cls,
        name: str,
        address: Address,
        id: Optional[UUID] = None,
        suites: Optional[List[Suite]] = None,
    ) -> "Building":
        if not id:
            id = uuid4()

        return cls(
            id,
            name,
            address,
            suites or [],
        )

    @property
    def suites(self) -> List[Suite]:
        return [v for v in self._suites.values()]

    def _get_suite(self, suite: Suite) -> Suite:
        _suite = self._suites.get(suite.number)

        if not _suite:
            raise ValueError(f"Suite {suite!r} does not exist")

        return _suite

    def add_suite(self, suite: Suite):
        if suite.number not in self._suites:
            self._suites[suite.number] = suite

            self._add_event(SuiteAdded(building_id=self._id, number=suite.number))
        else:
            raise ConstraintViolation("Suite already exists; update if need be")

    def lease_suite(self, suite: Suite):
        _suite = self._get_suite(suite)
        _suite.lease()

        self._add_event(SuiteLeased(building_id=self._id, number=suite.number))

    def make_suite_available(self, suite: Suite):
        _suite = self._get_suite(suite)
        _suite.make_available()

        self._add_event(SuiteMadeAvailable(building_id=self._id, number=suite.number))

    def remove_suite(self, suite: Suite):
        if suite.number not in self._suites:
            raise ValueError(f"Suite {suite!r} does not exist")

        suite_data = self._suites[suite.number].dict()
        del self._suites[suite.number]

        self._add_event(
            SuiteRemoved(
                building_id=self._id,
                name=suite_data["name"],
                number=suite_data["number"],
            )
        )

    def dict(self) -> NormalDict:
        return {
            "id": self._id,
            "address": self._address,
            "name": self._name,
            "suites": self.suites,
        }


def test_aggregate_root():
    # Create instance of building and its child entities/VOs
    address = Address(
        street_line_1="30 Rockefeller Plaza",
        locality="New York",
        region="NY",
        postal_code="10111",
        country_code_iso_alpha_3="USA",
    )

    suite_a = Suite.init("6700", "Top of the Rock", 67, 755602, True)
    suite_b = Suite.init("1280", "Suite 1280", 12, 1995, True)

    thirty_rock = Building.init(
        "30 Rockefeller Center",
        address,
        suites=[suite_a, suite_b],
    )

    # Test domain events
    with pytest.raises(ConstraintViolation):
        thirty_rock.lease_suite(suite_a)

    thirty_rock.make_suite_available(suite_b)

    suite_c = Suite.init("1735", "Suite 1735", 17, 1735, False)
    thirty_rock.add_suite(suite_c)

    thirty_rock.lease_suite(suite_c)

    suite_d = Suite.init("109", "Lobby Concessions", 1, 144, False)
    thirty_rock.add_suite(suite_d)

    thirty_rock.remove_suite(suite_d)

    # We expect five domain events
    event_1 = next(thirty_rock.events)
    event_2 = next(thirty_rock.events)
    event_3 = next(thirty_rock.events)
    event_4 = next(thirty_rock.events)
    event_5 = next(thirty_rock.events)

    event_5_data = event_5.dict()

    assert event_5_data["name"] == "Lobby Concessions"
    assert event_5_data["number"] == "109"

    # Now there shouldn't be any more events

    with pytest.raises(StopIteration):
        next(thirty_rock.events)

    assert isinstance(event_1, SuiteMadeAvailable)
    assert isinstance(event_2, SuiteAdded)
    assert isinstance(event_3, SuiteLeased)
    assert isinstance(event_4, SuiteAdded)
    assert isinstance(event_5, SuiteRemoved)


# Sample entities and tests
# Test basic entity functionality sans AggregateRoot


class Sensor(Entity):
    def __init__(self, id: UUID, name: str, installed_on: datetime):
        self._id = id
        self._name = name
        self._installed_on = installed_on

    @classmethod
    def init(
        cls,
        name: str,
        installed_on: datetime,
        id: Optional[UUID] = None,
    ) -> "Sensor":
        if not id:
            id = uuid4()

        return cls(id, name, installed_on)

    def dict(self) -> NormalDict:
        return {
            "id": self._id,
            "name": self._name,
            "installed_on": self._installed_on,
        }


class User(Entity):
    def __init__(self, id: int, name: str, is_active: bool):
        super().__init__()
        self._id = id
        self._is_active = is_active
        self._name = name

    @classmethod
    def init(
        cls, name: str, id: Optional[int] = None, is_active: Optional[bool] = True
    ) -> "User":
        if not id:
            id = randint(100, 10000)

        return cls(id, name, is_active)

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    def dict(self) -> NormalDict:
        return {"id": self._id, "name": self._name}


class Vehicle(Entity):
    def __init__(self, color: str):
        super().__init__()
        self._color = color

    @classmethod
    def init(cls, color: str) -> "Vehicle":
        return cls(color)

    def dict(self) -> NormalDict:
        return {
            "color": self._color,
        }


def test_entity_eq():
    assert Vehicle.init("green") == Vehicle.init("green")


def test_entity_hash_uuid_datetime():
    sensor = Sensor.init(
        "Humidity",
        datetime.now(),
    )

    # This checks that the json serialize used by Entity.__hash__ supports
    # datetime.datetime and UUID
    assert isinstance(hash(sensor), int)


def test_entity_repr():
    user = User.init("Gemma G")
    id = user.id

    assert user.__repr__() == f"User(id={id}, name='Gemma G')"


def test_entity_repr_no_properties():
    vehicle = Vehicle.init("red")

    assert vehicle.__repr__() == f"Vehicle()"


# Test using a custom serializer for hashing
# A custom serializer is necessary for types not supported by `orjson`


class Astronaut(Entity):
    def __init__(self, id: UUID, name: str):
        self._id = id
        self._name = name

    @classmethod
    def init(cls, name: str, id: Optional[UUID] = None) -> "Astronaut":
        if not id:
            id = uuid4()

        return cls(id, name)

    def dict(self) -> NormalDict:
        return {
            "id": self._id,
            "name": self._name,
        }


class RocketType:
    def __init__(self, name: str):
        if name not in ["falcon_9", "falcon_heavy", "starship"]:
            raise ValueError(f"Unknown rocket type {name!r}")

        self._name = name

    def __str__(self) -> str:
        return self._name


class SpaceDomainSerializer:
    def __call__(self, obj: Any) -> Any:
        if isinstance(obj, Astronaut):
            return obj.dict()
        elif isinstance(obj, RocketType):
            return str(obj)

        raise TypeError


class Rocket(AggregateRoot, Entity):
    def __init__(self, name: str, rocket_type: RocketType, astronauts: List[Astronaut]):
        super().__init__()
        self._astronauts = astronauts
        self._name = name
        self._rocket_type = rocket_type

    @classmethod
    def init(
        cls,
        name: str,
        rocket_type: RocketType,
        astronauts: Optional[List[Astronaut]] = None,
    ) -> "Rocket":
        return cls(
            name,
            rocket_type,
            astronauts or [],
        )

    def dict(self) -> NormalDict:
        return {
            "name": self._name,
            "rocket_type": self._rocket_type,
            "astronauts": self._astronauts,
        }


class SpaceShip(AggregateRoot, Entity, default_json_serializer=SpaceDomainSerializer):
    def __init__(self, name: str, rocket_type: RocketType, astronauts: List[Astronaut]):
        super().__init__()
        self._astronauts = astronauts
        self._name = name
        self._rocket_type = rocket_type

    @classmethod
    def init(
        cls,
        name: str,
        rocket_type: RocketType,
        astronauts: Optional[List[Astronaut]] = None,
    ) -> "SpaceShip":
        return cls(
            name,
            rocket_type,
            astronauts or [],
        )

    def dict(self) -> NormalDict:
        return {
            "name": self._name,
            "rocket_type": self._rocket_type,
            "astronauts": self._astronauts,
        }


def _set_up_deps() -> Tuple[Astronaut, RocketType]:
    astronaut = Astronaut.init("Albert II")
    rocket_type = RocketType("falcon_9")

    return astronaut, rocket_type


def test_entity_hash_custom_serializer():
    astronaut, rocket_type = _set_up_deps()
    space_ship = SpaceShip.init("Explorer", rocket_type, [astronaut])

    json = (
        '{"name":"Explorer","rocket_type":"falcon_9",'
        '"astronauts":[{"id":"%s","name":"Albert II"}]}'
    ) % (str(astronaut.dict()["id"]))

    assert space_ship.json().decode("utf-8") == json
    assert isinstance(hash(space_ship), int)


def test_entity_no_custom_serializer_type_error():
    """Not using a custom serializer will raise a type error upon trying to hash"""
    astronaut, rocket_type = _set_up_deps()
    rocket = Rocket.init("Explorer", rocket_type, [astronaut])

    with pytest.raises(TypeError):
        hash(rocket)
